# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
import json
import numpy
from tvb.simulator.models import ModelsEnum
from tvb.core.entities.file.simulator.configurations_h5 import SimulatorConfigurationH5
from tvb.core.neotraits.h5 import DataSet, Json, JsonFinal


class StateVariablesEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, numpy.ndarray):
            o = o.tolist()
        return o


class StateVariablesDecoder(json.JSONDecoder):
    def __init__(self):
        json.JSONDecoder.__init__(self, object_hook=self.dict_array)

    def dict_array(self, dictionary):
        dict_array = {}
        for k, v in dictionary.items():
            dict_array.update({k: numpy.array(v)})
        return dict_array


class EpileptorH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(EpileptorH5, self).__init__(path)

        self.a = DataSet(ModelsEnum.EPILEPTOR.get_class().a, self)
        self.b = DataSet(ModelsEnum.EPILEPTOR.get_class().b, self)
        self.c = DataSet(ModelsEnum.EPILEPTOR.get_class().c, self)
        self.d = DataSet(ModelsEnum.EPILEPTOR.get_class().d, self)
        self.r = DataSet(ModelsEnum.EPILEPTOR.get_class().r, self)
        self.s = DataSet(ModelsEnum.EPILEPTOR.get_class().s, self)
        self.x0 = DataSet(ModelsEnum.EPILEPTOR.get_class().x0, self)
        self.Iext = DataSet(ModelsEnum.EPILEPTOR.get_class().Iext, self)
        self.slope = DataSet(ModelsEnum.EPILEPTOR.get_class().slope, self)
        self.Iext2 = DataSet(ModelsEnum.EPILEPTOR.get_class().Iext2, self)
        self.tau = DataSet(ModelsEnum.EPILEPTOR.get_class().tau, self)
        self.aa = DataSet(ModelsEnum.EPILEPTOR.get_class().aa, self)
        self.bb = DataSet(ModelsEnum.EPILEPTOR.get_class().bb, self)
        self.Kvf = DataSet(ModelsEnum.EPILEPTOR.get_class().Kvf, self)
        self.Kf = DataSet(ModelsEnum.EPILEPTOR.get_class().Kf, self)
        self.Ks = DataSet(ModelsEnum.EPILEPTOR.get_class().Ks, self)
        self.tt = DataSet(ModelsEnum.EPILEPTOR.get_class().tt, self)
        self.modification = DataSet(ModelsEnum.EPILEPTOR.get_class().modification, self)
        self.state_variable_range = JsonFinal(ModelsEnum.EPILEPTOR.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.EPILEPTOR.get_class().variables_of_interest, self)


class Epileptor2DH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(Epileptor2DH5, self).__init__(path)

        self.a = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().a, self)
        self.b = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().b, self)
        self.c = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().c, self)
        self.d = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().d, self)
        self.r = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().r, self)
        self.x0 = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().x0, self)
        self.Iext = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().Iext, self)
        self.slope = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().slope, self)
        self.Kvf = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().Kvf, self)
        self.Ks = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().Ks, self)
        self.tt = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().tt, self)
        self.modification = DataSet(ModelsEnum.EPILEPTOR_2D.get_class().modification, self)
        self.state_variable_range = JsonFinal(ModelsEnum.EPILEPTOR_2D.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.EPILEPTOR_2D.get_class().variables_of_interest, self)


class EpileptorCodim3H5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(EpileptorCodim3H5, self).__init__(path)
        self.mu1_start = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().mu1_start, self)
        self.mu2_start = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().mu2_start, self)
        self.nu_start = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().nu_start, self)
        self.mu1_stop = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().mu1_stop, self)
        self.mu2_stop = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().mu2_stop, self)
        self.nu_stop = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().nu_stop, self)
        self.b = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().b, self)
        self.R = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().R, self)
        self.c = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().c, self)
        self.dstar = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().dstar, self)
        self.Ks = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().Ks, self)
        self.N = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().N, self)
        self.modification = DataSet(ModelsEnum.EPILEPTOR_CODIM_3.get_class().modification, self)
        self.state_variable_range = JsonFinal(ModelsEnum.EPILEPTOR_CODIM_3.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.EPILEPTOR_CODIM_3.get_class().variables_of_interest, self)


class EpileptorCodim3SlowModH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(EpileptorCodim3SlowModH5, self).__init__(path)
        self.mu1_Ain = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu1_Ain, self)
        self.mu2_Ain = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu2_Ain, self)
        self.nu_Ain = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().nu_Ain, self)
        self.mu1_Bin = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu1_Bin, self)
        self.mu2_Bin = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu2_Bin, self)
        self.nu_Bin = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().nu_Bin, self)
        self.mu1_Aend = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu1_Aend, self)
        self.mu2_Aend = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu2_Aend, self)
        self.nu_Aend = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().nu_Aend, self)
        self.mu1_Bend = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu1_Bend, self)
        self.mu2_Bend = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().mu2_Bend, self)
        self.nu_Bend = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().nu_Bend, self)
        self.b = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().b, self)
        self.R = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().R, self)
        self.c = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().c, self)
        self.cA = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().cA, self)
        self.cB = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().cB, self)
        self.dstar = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().dstar, self)
        self.Ks = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().Ks, self)
        self.N = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().N, self)
        self.modification = DataSet(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().modification, self)
        self.state_variable_range = JsonFinal(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.EPILEPTOR_CODIM_3_SLOW.get_class().variables_of_interest, self)


class HopfieldH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(HopfieldH5, self).__init__(path)
        self.taux = DataSet(ModelsEnum.HOPFIELD.get_class().taux, self)
        self.tauT = DataSet(ModelsEnum.HOPFIELD.get_class().tauT, self)
        self.dynamic = DataSet(ModelsEnum.HOPFIELD.get_class().dynamic, self)
        self.state_variable_range = JsonFinal(ModelsEnum.HOPFIELD.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.HOPFIELD.get_class().variables_of_interest, self)


class JansenRitH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(JansenRitH5, self).__init__(path)
        self.A = DataSet(ModelsEnum.JANSEN_RIT.get_class().A, self)
        self.B = DataSet(ModelsEnum.JANSEN_RIT.get_class().B, self)
        self.a = DataSet(ModelsEnum.JANSEN_RIT.get_class().a, self)
        self.b = DataSet(ModelsEnum.JANSEN_RIT.get_class().b, self)
        self.v0 = DataSet(ModelsEnum.JANSEN_RIT.get_class().v0, self)
        self.nu_max = DataSet(ModelsEnum.JANSEN_RIT.get_class().nu_max, self)
        self.r = DataSet(ModelsEnum.JANSEN_RIT.get_class().r, self)
        self.J = DataSet(ModelsEnum.JANSEN_RIT.get_class().J, self)
        self.a_1 = DataSet(ModelsEnum.JANSEN_RIT.get_class().a_1, self)
        self.a_2 = DataSet(ModelsEnum.JANSEN_RIT.get_class().a_2, self)
        self.a_3 = DataSet(ModelsEnum.JANSEN_RIT.get_class().a_3, self)
        self.a_4 = DataSet(ModelsEnum.JANSEN_RIT.get_class().a_4, self)
        self.p_min = DataSet(ModelsEnum.JANSEN_RIT.get_class().p_min, self)
        self.p_max = DataSet(ModelsEnum.JANSEN_RIT.get_class().p_max, self)
        self.mu = DataSet(ModelsEnum.JANSEN_RIT.get_class().mu, self)
        self.state_variable_range = JsonFinal(ModelsEnum.JANSEN_RIT.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.JANSEN_RIT.get_class().variables_of_interest, self)


class ZetterbergJansenH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ZetterbergJansenH5, self).__init__(path)
        self.He = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().He, self)
        self.Hi = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().Hi, self)
        self.ke = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().ke, self)
        self.ki = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().ki, self)
        self.e0 = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().e0, self)
        self.rho_2 = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().rho_2, self)
        self.rho_1 = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().rho_1, self)
        self.gamma_1 = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_1, self)
        self.gamma_2 = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_2, self)
        self.gamma_3 = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_3, self)
        self.gamma_4 = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_4, self)
        self.gamma_5 = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_5, self)
        self.gamma_1T = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_1T, self)
        self.gamma_2T = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_2T, self)
        self.gamma_3T = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().gamma_3T, self)
        self.P = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().P, self)
        self.U = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().U, self)
        self.Q = DataSet(ModelsEnum.ZETTERBERG_JANSEN.get_class().Q, self)
        self.state_variable_range = JsonFinal(ModelsEnum.ZETTERBERG_JANSEN.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.ZETTERBERG_JANSEN.get_class().variables_of_interest, self)


class EpileptorRestingStateH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(EpileptorRestingStateH5, self).__init__(path)
        self.a = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().a, self)
        self.b = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().b, self)
        self.c = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().c, self)
        self.d = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().d, self)
        self.r = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().r, self)
        self.s = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().s, self)
        self.x0 = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().x0, self)
        self.Iext = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().Iext, self)
        self.slope = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().slope, self)
        self.Iext2 = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().Iext2, self)
        self.tau = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().tau, self)
        self.aa = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().aa, self)
        self.bb = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().bb, self)
        self.Kvf = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().Kvf, self)
        self.Kf = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().Kf, self)
        self.Ks = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().Ks, self)
        self.tt = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().tt, self)
        self.tau_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().tau_rs, self)
        self.I_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().I_rs, self)
        self.a_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().a_rs, self)
        self.b_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().b_rs, self)
        self.d_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().d_rs, self)
        self.e_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().e_rs, self)
        self.f_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().f_rs, self)
        self.alpha_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().alpha_rs, self)
        self.beta_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().beta_rs, self)
        self.gamma_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().gamma_rs, self)
        self.K_rs = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().K_rs, self)
        self.p = DataSet(ModelsEnum.EPILEPTOR_RS.get_class().p, self)
        self.state_variable_range = JsonFinal(ModelsEnum.EPILEPTOR_RS.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.EPILEPTOR_RS.get_class().variables_of_interest, self)


class LarterBreakspearH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(LarterBreakspearH5, self).__init__(path)
        self.gCa = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().gCa, self)
        self.gK = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().gK, self)
        self.gL = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().gL, self)
        self.phi = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().phi, self)
        self.gNa = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().gNa, self)
        self.TK = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().TK, self)
        self.TCa = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().TCa, self)
        self.TNa = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().TNa, self)
        self.VCa = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().VCa, self)
        self.VK = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().VK, self)
        self.VL = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().VL, self)
        self.VNa = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().VNa, self)
        self.d_K = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_K, self)
        self.tau_K = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().tau_K, self)
        self.d_Na = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_Na, self)
        self.d_Ca = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_Ca, self)
        self.aei = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().aei, self)
        self.aie = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().aie, self)
        self.b = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().b, self)
        self.C = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().C, self)
        self.ane = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().ane, self)
        self.ani = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().ani, self)
        self.aee = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().aee, self)
        self.Iext = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().Iext, self)
        self.rNMDA = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().rNMDA, self)
        self.VT = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().VT, self)
        self.d_V = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_V, self)
        self.ZT = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().ZT, self)
        self.d_Z = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().d_Z, self)
        self.QV_max = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().QV_max, self)
        self.QZ_max = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().QZ_max, self)
        self.t_scale = DataSet(ModelsEnum.LARTER_BREAKSPEAR.get_class().t_scale, self)
        self.variables_of_interest = Json(ModelsEnum.LARTER_BREAKSPEAR.get_class().variables_of_interest, self)
        self.state_variable_range = JsonFinal(ModelsEnum.LARTER_BREAKSPEAR.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)


class LinearH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(LinearH5, self).__init__(path)
        self.gamma = DataSet(ModelsEnum.LINEAR.get_class().gamma, self)
        self.state_variable_range = JsonFinal(ModelsEnum.LINEAR.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.LINEAR.get_class().variables_of_interest, self)


class Generic2dOscillatorH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(Generic2dOscillatorH5, self).__init__(path)
        self.tau = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().tau, self)
        self.I = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().I, self)
        self.a = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().a, self)
        self.b = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().b, self)
        self.c = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().c, self)
        self.d = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().d, self)
        self.e = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().e, self)
        self.f = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().f, self)
        self.g = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().g, self)
        self.alpha = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().alpha, self)
        self.beta = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().beta, self)
        self.gamma = DataSet(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().gamma, self)
        self.state_variable_range = JsonFinal(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.GENERIC_2D_OSCILLATOR.get_class().variables_of_interest, self)


class KuramotoH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(KuramotoH5, self).__init__(path)
        self.omega = DataSet(ModelsEnum.KURAMOTO.get_class().omega, self)
        self.state_variable_range = JsonFinal(ModelsEnum.KURAMOTO.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.KURAMOTO.get_class().variables_of_interest, self)


class SupHopfH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(SupHopfH5, self).__init__(path)
        self.a = DataSet(ModelsEnum.SUP_HOPF.get_class().a, self)
        self.omega = DataSet(ModelsEnum.SUP_HOPF.get_class().omega, self)
        self.state_variable_range = JsonFinal(ModelsEnum.SUP_HOPF.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.SUP_HOPF.get_class().variables_of_interest, self)


class ReducedSetFitzHughNagumoH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ReducedSetFitzHughNagumoH5, self).__init__(path)
        self.tau = DataSet(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().tau, self)
        self.a = DataSet(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().a, self)
        self.b = DataSet(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().b, self)
        self.K11 = DataSet(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().K11, self)
        self.K12 = DataSet(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().K12, self)
        self.K21 = DataSet(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().K21, self)
        self.sigma = DataSet(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().sigma, self)
        self.mu = DataSet(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().mu, self)
        self.state_variable_range = JsonFinal(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().state_variable_range,
                                              self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.REDUCED_SET_FITZ_HUGH_NAGUMO.get_class().variables_of_interest,
                                          self)


class ReducedSetHindmarshRoseH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ReducedSetHindmarshRoseH5, self).__init__(path)
        self.r = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().r, self)
        self.a = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().a, self)
        self.b = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().b, self)
        self.c = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().c, self)
        self.d = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().d, self)
        self.s = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().s, self)
        self.xo = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().xo, self)
        self.K11 = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().K11, self)
        self.K12 = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().K12, self)
        self.K21 = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().K21, self)
        self.sigma = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().sigma, self)
        self.mu = DataSet(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().mu, self)
        self.state_variable_range = JsonFinal(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().state_variable_range,
                                              self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.REDUCED_SET_HINDMARSH_ROSE.get_class().variables_of_interest, self)


class WilsonCowanH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(WilsonCowanH5, self).__init__(path)
        self.c_ee = DataSet(ModelsEnum.WILSON_COWAN.get_class().c_ee, self)
        self.c_ie = DataSet(ModelsEnum.WILSON_COWAN.get_class().c_ie, self)
        self.c_ei = DataSet(ModelsEnum.WILSON_COWAN.get_class().c_ei, self)
        self.c_ii = DataSet(ModelsEnum.WILSON_COWAN.get_class().c_ii, self)
        self.tau_e = DataSet(ModelsEnum.WILSON_COWAN.get_class().tau_e, self)
        self.tau_i = DataSet(ModelsEnum.WILSON_COWAN.get_class().tau_i, self)
        self.a_e = DataSet(ModelsEnum.WILSON_COWAN.get_class().a_e, self)
        self.b_e = DataSet(ModelsEnum.WILSON_COWAN.get_class().b_e, self)
        self.c_e = DataSet(ModelsEnum.WILSON_COWAN.get_class().c_e, self)
        self.theta_e = DataSet(ModelsEnum.WILSON_COWAN.get_class().theta_e, self)
        self.a_i = DataSet(ModelsEnum.WILSON_COWAN.get_class().a_i, self)
        self.b_i = DataSet(ModelsEnum.WILSON_COWAN.get_class().b_i, self)
        self.theta_i = DataSet(ModelsEnum.WILSON_COWAN.get_class().theta_i, self)
        self.c_i = DataSet(ModelsEnum.WILSON_COWAN.get_class().c_i, self)
        self.r_e = DataSet(ModelsEnum.WILSON_COWAN.get_class().r_e, self)
        self.r_i = DataSet(ModelsEnum.WILSON_COWAN.get_class().r_i, self)
        self.k_e = DataSet(ModelsEnum.WILSON_COWAN.get_class().k_e, self)
        self.k_i = DataSet(ModelsEnum.WILSON_COWAN.get_class().k_i, self)
        self.P = DataSet(ModelsEnum.WILSON_COWAN.get_class().P, self)
        self.Q = DataSet(ModelsEnum.WILSON_COWAN.get_class().Q, self)
        self.alpha_e = DataSet(ModelsEnum.WILSON_COWAN.get_class().alpha_e, self)
        self.alpha_i = DataSet(ModelsEnum.WILSON_COWAN.get_class().alpha_i, self)
        self.state_variable_range = JsonFinal(ModelsEnum.WILSON_COWAN.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.WILSON_COWAN.get_class().variables_of_interest, self)


class ReducedWongWangH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ReducedWongWangH5, self).__init__(path)
        self.a = DataSet(ModelsEnum.REDUCED_WONG_WANG.get_class().a, self)
        self.b = DataSet(ModelsEnum.REDUCED_WONG_WANG.get_class().b, self)
        self.d = DataSet(ModelsEnum.REDUCED_WONG_WANG.get_class().d, self)
        self.gamma = DataSet(ModelsEnum.REDUCED_WONG_WANG.get_class().gamma, self)
        self.tau_s = DataSet(ModelsEnum.REDUCED_WONG_WANG.get_class().tau_s, self)
        self.w = DataSet(ModelsEnum.REDUCED_WONG_WANG.get_class().w, self)
        self.J_N = DataSet(ModelsEnum.REDUCED_WONG_WANG.get_class().J_N, self)
        self.I_o = DataSet(ModelsEnum.REDUCED_WONG_WANG.get_class().I_o, self)
        self.sigma_noise = DataSet(ModelsEnum.REDUCED_WONG_WANG.get_class().sigma_noise, self)
        self.state_variable_range = JsonFinal(ModelsEnum.REDUCED_WONG_WANG.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.REDUCED_WONG_WANG.get_class().variables_of_interest, self)


class ReducedWongWangExcInhH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ReducedWongWangExcInhH5, self).__init__(path)
        self.a_e = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().a_e, self)
        self.b_e = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().b_e, self)
        self.d_e = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().d_e, self)
        self.gamma_e = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().gamma_e, self)
        self.tau_e = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().tau_e, self)
        self.w_p = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().w_p, self)
        self.J_N = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().J_N, self)
        self.W_e = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().W_e, self)
        self.a_i = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().a_i, self)
        self.b_i = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().b_i, self)
        self.d_i = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().d_i, self)
        self.gamma_i = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().gamma_i, self)
        self.tau_i = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().tau_i, self)
        self.J_i = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().J_i, self)
        self.W_i = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().W_i, self)
        self.I_o = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().I_o, self)
        self.G = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().G, self)
        self.lamda = DataSet(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().lamda, self)
        self.state_variable_range = JsonFinal(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().state_variable_range,
                                              self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.REDUCED_WONG_WANG_EXCH_INH.get_class().variables_of_interest, self)


class ZerlautAdaptationFirstOrderH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ZerlautAdaptationFirstOrderH5, self).__init__(path)
        self.g_L = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().g_L, self)
        self.E_L_e = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().E_L_e, self)
        self.E_L_i = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().E_L_i, self)
        self.C_m = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().C_m, self)
        self.a_e = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().a_e, self)
        self.a_i = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().a_i, self)
        self.b_e = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().b_e, self)
        self.b_i = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().b_i, self)
        self.tau_w_e = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().tau_w_e, self)
        self.tau_w_i = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().tau_w_i, self)
        self.E_e = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().E_e, self)
        self.E_i = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().E_i, self)
        self.Q_e = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().Q_e, self)
        self.Q_i = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().Q_i, self)
        self.tau_e = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().tau_e, self)
        self.tau_i = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().tau_i, self)
        self.N_tot = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().N_tot, self)
        self.p_connect = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().p_connect, self)
        self.g = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().g, self)
        self.K_ext_e = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().K_ext_e, self)
        self.K_ext_i = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().K_ext_i, self)
        self.T = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().T, self)
        self.P_e = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().P_e, self)
        self.P_i = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().P_i, self)
        self.external_input_ex_ex = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().external_input_ex_ex, self)
        self.external_input_ex_in = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().external_input_ex_in, self)
        self.external_input_in_ex = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().external_input_in_ex, self)
        self.external_input_in_in = DataSet(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().external_input_in_in, self)
        self.state_variable_range = JsonFinal(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.ZERLAUT_FIRST_ORDER.get_class().variables_of_interest, self)


class ZerlautAdaptationSecondOrderH5(ZerlautAdaptationFirstOrderH5):

    def __init__(self, path):
        super(ZerlautAdaptationSecondOrderH5, self).__init__(path)
        self.state_variable_range = JsonFinal(ModelsEnum.ZERLAUT_SECOND_ORDER.get_class().state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ModelsEnum.ZERLAUT_SECOND_ORDER.get_class().variables_of_interest, self)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Adapter that uses the traits module to generate interfaces for FFT Analyzer.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Paula Sanz Leon <paula@tvb.invalid>

"""
import uuid
import numpy
from tvb.analyzers.node_complex_coherence import NodeComplexCoherence
from tvb.core.adapters.abcadapter import ABCAsynchronous, ABCAdapterForm
from tvb.core.neotraits.view_model import ViewModel, DataTypeGidAttr
from tvb.datatypes.time_series import TimeSeries
from tvb.core.entities.filters.chain import FilterChain
from tvb.adapters.datatypes.h5.spectral_h5 import ComplexCoherenceSpectrumH5
from tvb.adapters.datatypes.db.spectral import ComplexCoherenceSpectrumIndex
from tvb.adapters.datatypes.db.time_series import TimeSeriesIndex
from tvb.core.neotraits.forms import TraitDataTypeSelectField
from tvb.core.neocom import h5


class NodeComplexCoherenceModel(ViewModel, NodeComplexCoherence):
    time_series = DataTypeGidAttr(
        linked_datatype=TimeSeries,
        label="Time Series",
        required=True,
        doc="""The timeseries for which the CrossCoherence and ComplexCoherence is to be computed."""
    )


class NodeComplexCoherenceForm(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None):
        super(NodeComplexCoherenceForm, self).__init__(prefix, project_id)
        self.time_series = TraitDataTypeSelectField(NodeComplexCoherenceModel.time_series, self,
                                                    name=self.get_input_name(),
                                                    conditions=self.get_filters())

    @staticmethod
    def get_view_model():
        return NodeComplexCoherenceModel

    @staticmethod
    def get_required_datatype():
        return TimeSeriesIndex

    @staticmethod
    def get_filters():
        return FilterChain(fields=[FilterChain.datatype + '.data_ndim'], operations=["=="], values=[4])

    @staticmethod
    def get_input_name():
        return "time_series"

    def get_traited_datatype(self):
        return NodeComplexCoherence()


class NodeComplexCoherenceAdapter(ABCAsynchronous):
    """ TVB adapter for calling the NodeComplexCoherence algorithm. """

    _ui_name = "Complex Coherence of Nodes"
    _ui_description = "Compute the node complex (imaginary) coherence for a TimeSeries input DataType."
    _ui_subsection = "complexcoherence"

    def get_form_class(self):
        return NodeComplexCoherenceForm

    def get_output(self):
        return [ComplexCoherenceSpectrumIndex]

    def get_required_memory_size(self, view_model):
        # type: (NodeComplexCoherenceModel) -> int
        """
        Return the required memory to run this algorithm.
        """
        input_size = numpy.prod(self.input_shape) * 8.0
        output_size = self.algorithm.result_size(self.input_shape, self.algorithm.max_freq,
                                                 self.algorithm.epoch_length,
                                                 self.algorithm.segment_length,
                                                 self.algorithm.segment_shift,
                                                 self.input_time_series_index.sample_period,
                                                 self.algorithm.zeropad,
                                                 self.algorithm.average_segments)

        return input_size + output_size

    def get_required_disk_size(self, view_model):
        # type: (NodeComplexCoherenceModel) -> int
        """
        Returns the required disk size to be able to run the adapter (in kB).
        """
        result = self.algorithm.result_size(self.input_shape, self.algorithm.max_freq,
                                            self.algorithm.epoch_length,
                                            self.algorithm.segment_length,
                                            self.algorithm.segment_shift,
                                            self.input_time_series_index.sample_period,
                                            self.algorithm.zeropad,
                                            self.algorithm.average_segments)
        return self.array_size2kb(result)

    def configure(self, view_model):
        # type: (NodeComplexCoherenceModel) -> None
        """
        Do any configuration needed before launching and create an instance of the algorithm.
        """
        self.input_time_series_index = self.load_entity_by_gid(view_model.time_series.hex)
        self.input_shape = (self.input_time_series_index.data_length_1d,
                            self.input_time_series_index.data_length_2d,
                            self.input_time_series_index.data_length_3d,
                            self.input_time_series_index.data_length_4d)
        self.log.debug("Time series shape is %s" % (str(self.input_shape)))
        # -------------------- Fill Algorithm for Analysis -------------------##
        self.algorithm = NodeComplexCoherence()
        self.memory_factor = 1

    def launch(self, view_model):
        # type: (NodeComplexCoherenceModel) -> [ComplexCoherenceSpectrumIndex]
        """
        Launch algorithm and build results.

        :returns: the `ComplexCoherenceSpectrum` built with the given time-series
        """
        # ------- Prepare a ComplexCoherenceSpectrum object for result -------##
        complex_coherence_spectrum_index = ComplexCoherenceSpectrumIndex()
        time_series_h5 = h5.h5_file_for_index(self.input_time_series_index)

        dest_path = h5.path_for(self.storage_path, ComplexCoherenceSpectrumH5, complex_coherence_spectrum_index.gid)
        spectra_h5 = ComplexCoherenceSpectrumH5(dest_path)
        spectra_h5.gid.store(uuid.UUID(complex_coherence_spectrum_index.gid))
        spectra_h5.source.store(time_series_h5.gid.load())

        # ------------------- NOTE: Assumes 4D TimeSeries. -------------------##
        input_shape = time_series_h5.data.shape
        node_slice = [slice(input_shape[0]), slice(input_shape[1]), slice(input_shape[2]), slice(input_shape[3])]

        # ---------- Iterate over slices and compose final result ------------##
        small_ts = TimeSeries()
        small_ts.sample_period = time_series_h5.sample_period.load()
        small_ts.data = time_series_h5.read_data_slice(tuple(node_slice))
        self.algorithm.time_series = small_ts

        partial_result = self.algorithm.evaluate()
        self.log.debug("got partial_result")
        self.log.debug("partial segment_length is %s" % (str(partial_result.segment_length)))
        self.log.debug("partial epoch_length is %s" % (str(partial_result.epoch_length)))
        self.log.debug("partial windowing_function is %s" % (str(partial_result.windowing_function)))
        # LOG.debug("partial frequency vector is %s" % (str(partial_result.frequency)))

        spectra_h5.write_data_slice(partial_result)
        spectra_h5.segment_length.store(partial_result.segment_length)
        spectra_h5.epoch_length.store(partial_result.epoch_length)
        spectra_h5.windowing_function.store(partial_result.windowing_function)
        # spectra.frequency = partial_result.frequency
        spectra_h5.close()
        time_series_h5.close()

        complex_coherence_spectrum_index.source_gid = self.input_time_series_index.gid
        complex_coherence_spectrum_index.epoch_length = partial_result.epoch_length
        complex_coherence_spectrum_index.segment_length = partial_result.segment_length
        complex_coherence_spectrum_index.windowing_function = partial_result.windowing_function
        complex_coherence_spectrum_index.frequency_step = partial_result.freq_step
        complex_coherence_spectrum_index.max_frequency = partial_result.max_freq

        return complex_coherence_spectrum_index



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

import uuid
import formencode
from formencode import validators
from tvb.core.entities.filters.chain import FilterChain
from tvb.basic.neotraits.api import Attr, Range
from tvb.datatypes.cortex import Cortex
from tvb.simulator.simulator import Simulator
from tvb.adapters.simulator.integrator_forms import get_ui_name_to_integrator_dict
from tvb.adapters.simulator.model_forms import get_ui_name_to_model
from tvb.adapters.simulator.monitor_forms import get_ui_name_to_monitor_dict
from tvb.adapters.simulator.range_parameter import RangeParameter
from tvb.core.adapters.abcadapter import ABCAdapterForm
from tvb.adapters.datatypes.db.local_connectivity import LocalConnectivityIndex
from tvb.adapters.datatypes.db.patterns import StimuliSurfaceIndex, StimuliRegionIndex
from tvb.adapters.datatypes.db.region_mapping import RegionMappingIndex
from tvb.adapters.datatypes.db.surface import SurfaceIndex
from tvb.core.neotraits.forms import DataTypeSelectField, SimpleSelectField, ScalarField, ArrayField, SimpleFloatField, \
    SimpleHiddenField
from tvb.core.neocom import h5


class SimulatorSurfaceFragment(ABCAdapterForm):
    def __init__(self, prefix='', project_id=None):
        super(SimulatorSurfaceFragment, self).__init__(prefix, project_id)
        # TODO: should show only corticals
        self.surface = DataTypeSelectField(SurfaceIndex, self, name='surface', required=False,
                                           label=Simulator.surface.label, doc=Simulator.surface.doc)

    def fill_from_trait(self, trait):
        # type: (Simulator) -> None
        if trait.surface:
            if hasattr(trait.surface, 'surface_gid'):
                self.surface.data = trait.surface.surface_gid.hex
        else:
            self.surface.data = None


class SimulatorRMFragment(ABCAdapterForm):
    def __init__(self, prefix='', project_id=None, surface_index=None):
        super(SimulatorRMFragment, self).__init__(prefix, project_id)
        conditions = None
        if surface_index:
            conditions = FilterChain(fields=[FilterChain.datatype + '.surface_gid'], operations=["=="],
                                     values=[str(surface_index.gid)])
        self.rm = DataTypeSelectField(RegionMappingIndex, self, name='region_mapping', required=True,
                                      label=Cortex.region_mapping_data.label,
                                      doc=Cortex.region_mapping_data.doc, conditions=conditions)
        self.lc = DataTypeSelectField(LocalConnectivityIndex, self, name='local_connectivity',
                                      label=Cortex.local_connectivity.label, doc=Cortex.local_connectivity.doc,
                                      conditions=conditions)
        self.coupling_strength = ArrayField(Cortex.coupling_strength, self)

    def fill_from_trait(self, trait):
        # type: (Simulator) -> None
        self.coupling_strength.data = trait.surface.coupling_strength
        if hasattr(trait.surface, 'region_mapping_data'):
            self.rm.data = trait.surface.region_mapping_data.hex
        else:
            self.rm.data = None
        if trait.surface.local_connectivity:
            self.lc.data = trait.surface.local_connectivity.hex
        else:
            self.lc.data = None


class SimulatorStimulusFragment(ABCAdapterForm):
    def __init__(self, prefix='', project_id=None, is_surface_simulation=False):
        super(SimulatorStimulusFragment, self).__init__(prefix, project_id)
        if is_surface_simulation:
            stimuli_index_class = StimuliSurfaceIndex
        else:
            stimuli_index_class = StimuliRegionIndex
        self.stimulus = DataTypeSelectField(stimuli_index_class, self, name='region_stimuli', required=False,
                                            label=Simulator.stimulus.label, doc=Simulator.stimulus.doc)

    def fill_from_trait(self, trait):
        # type: (Simulator) -> None
        if hasattr(trait, 'stimulus') and trait.stimulus is not None:
            self.stimulus.data = trait.stimulus.hex
        else:
            self.stimulus.data = None


class SimulatorModelFragment(ABCAdapterForm):
    def __init__(self, prefix='', project_id=None):
        super(SimulatorModelFragment, self).__init__(prefix, project_id)

        self.model_choices = get_ui_name_to_model()

        self.model = SimpleSelectField(choices=self.model_choices, form=self, name='model', required=True,
                                       label=Simulator.model.label, doc=Simulator.model.doc)
        self.model.template = "form_fields/select_field.html"

    def fill_from_trait(self, trait):
        # type: (Simulator) -> None
        self.model.data = trait.model.__class__


class SimulatorIntegratorFragment(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None):
        super(SimulatorIntegratorFragment, self).__init__(prefix, project_id)

        self.integrator_choices = get_ui_name_to_integrator_dict()

        self.integrator = SimpleSelectField(choices=self.integrator_choices, form=self, name='integrator',
                                            required=True,
                                            label=Simulator.integrator.label, doc=Simulator.integrator.doc)
        self.integrator.template = "form_fields/select_field.html"

    def fill_from_trait(self, trait):
        # type: (Simulator) -> None
        self.integrator.data = trait.integrator.__class__


class SimulatorMonitorFragment(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None, is_surface_simulation=False):
        super(SimulatorMonitorFragment, self).__init__(prefix, project_id)

        self.monitor_choices = get_ui_name_to_monitor_dict(is_surface_simulation)

        self.monitor = SimpleSelectField(choices=self.monitor_choices, form=self, name='monitor', required=True,
                                         label=Simulator.monitors.label, doc=Simulator.monitors.doc)
        self.monitor.template = "form_fields/select_field.html"

    def fill_from_trait(self, trait):
        # type: (Simulator) -> None
        self.monitor.data = trait.monitors[0].__class__


class SimulatorLengthFragment(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None):
        super(SimulatorLengthFragment, self).__init__(prefix, project_id)
        self.length = ScalarField(Simulator.simulation_length, self)


class SimulatorFinalFragment(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None, simulation_number=1):
        super(SimulatorFinalFragment, self).__init__(prefix, project_id)
        default_simulation_name = "simulation_" + str(simulation_number)
        self.simulation_name = ScalarField(Attr(str, doc='Name for the current simulation configuration', default=default_simulation_name,
                                                label='Simulation name'), self, name='input-simulation-name-id')

    def fill_from_post(self, form_data):
        super(SimulatorFinalFragment, self).fill_from_post(form_data)
        valiadation_result = SimulatorFinalFragment.is_burst_name_ok(self.simulation_name.value)
        if valiadation_result is not True:
            raise ValueError(valiadation_result)

    @staticmethod
    def is_burst_name_ok(burst_name):
        class BurstNameForm(formencode.Schema):
            """
            Validate Burst name string
            """
            burst_name = formencode.All(validators.UnicodeString(not_empty=True),
                                        validators.Regex(regex=r"^[a-zA-Z\. _\-0-9]*$"))

        try:
            form = BurstNameForm()
            form.to_python({'burst_name': burst_name})
            return True
        except formencode.Invalid:
            validation_error = "Invalid simulation name %s. Please use only letters, numbers, or _ " % str(burst_name)
            return validation_error


class SimulatorPSEConfigurationFragment(ABCAdapterForm):

    def __init__(self, choices, prefix='', project_id=None):
        super(SimulatorPSEConfigurationFragment, self).__init__(prefix, project_id)
        self.pse_param1 = SimpleSelectField(choices, form=self, name='pse_param1', required=True, label="PSE param1")
        self.pse_param1.template = "form_fields/select_field.html"
        self.pse_param2 = SimpleSelectField(choices, form=self, name='pse_param2', label="PSE param2")
        self.pse_param2.template = "form_fields/select_field.html"


class SimulatorPSEParamRangeFragment(ABCAdapterForm):

    def __init__(self, pse_param1, pse_param2, prefix='', project_id=None):
        # type: (RangeParameter) -> None
        super(SimulatorPSEParamRangeFragment, self).__init__(prefix, project_id)
        self.pse_param1_name = SimpleHiddenField(self, name='pse_param1_name', default=pse_param1.name)
        if pse_param1.type is float:
            self.pse_param1_lo = SimpleFloatField(self, name='pse_param1_lo', required=True, label='pse_param1_lo',
                                                  default=pse_param1.range_definition.lo)
            self.pse_param1_hi = SimpleFloatField(self, name='pse_param1_hi', required=True, label='pse_param1_hi',
                                                  default=pse_param1.range_definition.hi)
            self.pse_param1_step = SimpleFloatField(self, name='pse_param1_step', required=True,
                                                    label='pse_param1_step', default=pse_param1.range_definition.step)

        else:
            self.pse_param1_dt = DataTypeSelectField(h5.REGISTRY.get_index_for_datatype(pse_param1.type), self,
                                                     name='pse_param1', required=True, label='pse_param1',
                                                     dynamic_conditions=pse_param1.range_definition,
                                                     has_all_option=True)

        if pse_param2:
            self.pse_param2_name = SimpleHiddenField(self, name='pse_param2_name', default=pse_param2.name)
            if pse_param2.type is float:
                self.pse_param2_lo = SimpleFloatField(self, name='pse_param2_lo', required=True, label='pse_param2_lo',
                                                      default=pse_param2.range_definition.lo)
                self.pse_param2_hi = SimpleFloatField(self, name='pse_param2_hi', required=True, label='pse_param2_hi',
                                                      default=pse_param2.range_definition.hi)
                self.pse_param2_step = SimpleFloatField(self, name='pse_param2_step', required=True,
                                                        label='pse_param2_step',
                                                        default=pse_param2.range_definition.step)
            else:
                self.pse_param2_dt = DataTypeSelectField(h5.REGISTRY.get_index_for_datatype(pse_param2.type), self,
                                                         name='pse_param2', required=True, label='pse_param2',
                                                         dynamic_conditions=pse_param2.range_definition,
                                                         has_all_option=True)

    @staticmethod
    def _prepare_pse_uuid_list(pse_uuid_str):
        pse_uuid_str_list = pse_uuid_str.split(',')
        pse_uuid_list = [uuid.UUID(uuid_str) for uuid_str in pse_uuid_str_list]
        return pse_uuid_list

    @staticmethod
    def fill_from_post(all_range_parameters, **data):
        pse_param1_name = data.get('_pse_param1_name')
        pse_param2_name = data.get('_pse_param2_name')

        pse_param1 = all_range_parameters.get(pse_param1_name)
        if pse_param1.type is float:
            pse_param1_lo = data.get('_pse_param1_lo')
            pse_param1_hi = data.get('_pse_param1_hi')
            pse_param1_step = data.get('_pse_param1_step')
            param1_range = RangeParameter(pse_param1_name, pse_param1.type,
                                          Range(float(pse_param1_lo), float(pse_param1_hi), float(pse_param1_step)),
                                          is_array=pse_param1.is_array)
        else:
            param1_range_str = data.get('_pse_param1')
            pse_param1_name += '.gid'
            param1_range = RangeParameter(pse_param1_name, pse_param1.type,
                                          SimulatorPSEParamRangeFragment._prepare_pse_uuid_list(param1_range_str))

        param2_range = None
        if pse_param2_name:
            pse_param2 = all_range_parameters.get(pse_param2_name)
            if pse_param2.type is float:
                pse_param2_lo = data.get('_pse_param2_lo')
                pse_param2_hi = data.get('_pse_param2_hi')
                pse_param2_step = data.get('_pse_param2_step')
                param2_range = RangeParameter(pse_param2_name, float,
                                              Range(float(pse_param2_lo), float(pse_param2_hi), float(pse_param2_step)),
                                              is_array=pse_param2.is_array)
            else:
                param2_range_str = data.get('_pse_param2')
                pse_param2_name += '.gid'
                param2_range = RangeParameter(pse_param2_name, pse_param1.type,
                                              SimulatorPSEParamRangeFragment._prepare_pse_uuid_list(param2_range_str))

        return param1_range, param2_range



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
# Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
# Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Change of DB structure from TVB version 1.4.1 to TVB 1.5

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

from sqlalchemy import Column, Integer, String, Boolean, DateTime
from sqlalchemy.sql import text
from migrate.changeset.constraint import ForeignKeyConstraint
from migrate.changeset.schema import create_column, drop_column
from tvb.basic.logger.builder import get_logger
from tvb.core.entities import model
from tvb.core.entities.storage import SA_SESSIONMAKER


meta = model.Base.metadata

LOGGER = get_logger(__name__)

DEL_COLUMNS = [Column('fk_algo_group', Integer),
               Column('identifier', String),
               Column('name', String)]

ADD_COLUMNS = [Column('fk_category', Integer),
               Column('module', String),
               Column('classname', String),
               Column('group_name', String),
               Column('group_description', String),
               Column('displayname', String),
               Column('subsection_name', String),
               Column('last_introspection_check', DateTime),
               Column('removed', Boolean, default=False)]



def upgrade(migrate_engine):
    """
    Alter existing table ALGORITHMS, by moving columns from the old ALGORITHM_GROUPS table.
    """
    meta.bind = migrate_engine
    table_algo = meta.tables["ALGORITHMS"]
    for col in ADD_COLUMNS:
        create_column(col, table_algo)

    session = SA_SESSIONMAKER()
    try:
        session.execute(text("ALTER TABLE \"MAPPED_SIMULATION_STATE\" "
                             "ADD COLUMN _current_state VARYING CHARACTER(255)"))
        session.commit()
    except Exception:
        session.close()
        session = SA_SESSIONMAKER()
        session.execute(text("ALTER TABLE \"MAPPED_SIMULATION_STATE\" "
                             "ADD COLUMN _current_state character varying;"))
        session.commit()
    finally:
        session.close()

    session = SA_SESSIONMAKER()
    try:
        session.execute(text(
            """UPDATE "ALGORITHMS" SET
            module = (select G.module FROM "ALGORITHM_GROUPS" G WHERE "ALGORITHMS".fk_algo_group=G.id),
            classname = (select G.classname FROM "ALGORITHM_GROUPS" G WHERE "ALGORITHMS".fk_algo_group=G.id),
            displayname = (select G.displayname FROM "ALGORITHM_GROUPS" G WHERE "ALGORITHMS".fk_algo_group=G.id),
            fk_category = (select G.fk_category FROM "ALGORITHM_GROUPS" G WHERE "ALGORITHMS".fk_algo_group=G.id);"""))
        session.commit()

        # Delete old columns, no longer needed
        for col in DEL_COLUMNS:
            drop_column(col, table_algo)

        # Create constraint only after rows are being populated
        table_algo = meta.tables["ALGORITHMS"]
        fk_constraint = ForeignKeyConstraint(["fk_category"], ["ALGORITHM_CATEGORIES.id"],
                                             ondelete="CASCADE", table=table_algo)
        fk_constraint.create()

        # Drop old table
        session = SA_SESSIONMAKER()
        session.execute(text("""DROP TABLE "ALGORITHM_GROUPS";"""))
        session.commit()

    except Exception as excep:
        LOGGER.exception(excep)
    finally:
        session.close()

    session = SA_SESSIONMAKER()
    try:
        session.execute(text("""ALTER TABLE "MAPPED_CONNECTIVITY_ANNOTATIONS"
                                RENAME TO "MAPPED_CONNECTIVITY_ANNOTATIONS_DATA"; """))
        session.execute(text("""ALTER TABLE "MAPPED_DATATYPE_MEASURE" RENAME TO "MAPPED_DATATYPE_MEASURE_DATA"; """))
        session.execute(text("""ALTER TABLE "MAPPED_SIMULATION_STATE" RENAME TO "MAPPED_SIMULATION_STATE_DATA"; """))
        session.execute(text("""ALTER TABLE "MAPPED_VALUE_WRAPPER" RENAME TO "MAPPED_VALUE_WRAPPER_DATA"; """))
        session.execute(text("""ALTER TABLE "MAPPED_PROJECTION_DATA" RENAME TO "MAPPED_PROJECTION_MATRIX_DATA"; """))
        session.commit()
    except Exception as excep:
        LOGGER.exception(excep)
    finally:
        session.close()


def downgrade(_):
    """
    Downgrade currently not supported
    """
    pass



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
from abc import abstractmethod
from tvb.basic.neotraits.api import NArray
from tvb.adapters.simulator.range_parameter import RangeParameter
from tvb.core.neotraits.forms import Form


class FormWithRanges(Form):
    @staticmethod
    @abstractmethod
    def get_params_configurable_in_phase_plane():
        """Return a list with all parameter names that can be configured in Phase Plane viewer for the current model"""

    def _get_parameters_for_pse(self):
        pass

    def _gather_parameters_with_range_defined(self):
        parameters_with_range = []
        for trait_field in self.trait_fields:
            if isinstance(trait_field.trait_attribute, NArray) and trait_field.trait_attribute.domain:
                parameters_with_range.append(trait_field.trait_attribute)
        return parameters_with_range

    def get_range_parameters(self):
        parameters_for_pse = self._gather_parameters_with_range_defined()

        result = []
        for range_param in parameters_for_pse:
            range_obj = RangeParameter(range_param.field_name, float, range_param.domain,
                                       isinstance(range_param, NArray))
            result.append(range_obj)
        return result



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

from datetime import datetime
from tvb.basic.logger.builder import get_logger
from tvb.core.entities.file.files_helper import FilesHelper
from tvb.core.entities.model.model_datatype import DataTypeGroup
from tvb.core.entities.model.model_burst import BurstConfiguration
from tvb.core.entities.storage import dao
from tvb.core.utils import format_bytes_human, format_timedelta

MAX_BURSTS_DISPLAYED = 50


class BurstService(object):

    def __init__(self):
        self.logger = get_logger(self.__class__.__module__)
        self.file_helper = FilesHelper()

    def mark_burst_finished(self, burst_entity, burst_status=None, error_message=None):
        """
        Mark Burst status field.
        Also compute 'weight' for current burst: no of operations inside, estimate time on disk...

        :param burst_entity: BurstConfiguration to be updated, at finish time.
        :param burst_status: BurstConfiguration status. By default BURST_FINISHED
        :param error_message: If given, set the status to error and perpetuate the message.
        """
        if burst_status is None:
            burst_status = BurstConfiguration.BURST_FINISHED
        if error_message is not None:
            burst_status = BurstConfiguration.BURST_ERROR

        try:
            # If there are any DataType Groups in current Burst, update their counter.
            burst_dt_groups = dao.get_generic_entity(DataTypeGroup, burst_entity.id, "fk_parent_burst")
            for dt_group in burst_dt_groups:
                dt_group.count_results = dao.count_datatypes_in_group(dt_group.id)
                dt_group.disk_size, dt_group.subject = dao.get_summary_for_group(dt_group.id)
                dao.store_entity(dt_group)

            # Update actual Burst entity fields
            burst_entity.datatypes_number = dao.count_datatypes_in_burst(burst_entity.id)

            burst_entity.status = burst_status
            burst_entity.error_message = error_message
            burst_entity.finish_time = datetime.now()
            dao.store_entity(burst_entity)
        except Exception:
            self.logger.exception("Could not correctly update Burst status and meta-data!")
            burst_entity.status = burst_status
            burst_entity.error_message = "Error when updating Burst Status"
            burst_entity.finish_time = datetime.now()
            dao.store_entity(burst_entity)

    def persist_operation_state(self, operation, operation_status, message=None):
        """
        Update Operation instance state. Store it in DB and on HDD/
        :param operation: Operation instance
        :param operation_status: new status
        :param message: message in case of error
        :return: operation instance changed
        """
        operation.mark_complete(operation_status, message)
        dao.store_entity(operation)
        operation = dao.get_operation_by_id(operation.id)
        self.file_helper.write_operation_metadata(operation)
        return operation

    def get_burst_for_operation_id(self, operation_id):
        return dao.get_burst_for_operation_id(operation_id)

    @staticmethod
    def rename_burst(burst_id, new_name):
        """
        Rename the burst given by burst_id, setting it's new name to
        burst_name.
        """
        burst = dao.get_burst_by_id(burst_id)
        burst.name = new_name
        dao.store_entity(burst)

    @staticmethod
    def get_available_bursts(project_id):
        """
        Return all the burst for the current project.
        """
        bursts = dao.get_bursts_for_project(project_id, page_size=MAX_BURSTS_DISPLAYED) or []
        # for burst in bursts:
        #     burst.prepare_after_load()
        return bursts

    @staticmethod
    def populate_burst_disk_usage(bursts):
        """
        Adds a disk_usage field to each burst object.
        The disk usage is computed as the sum of the datatypes generated by a burst
        """
        sizes = dao.compute_bursts_disk_size([b.id for b in bursts])
        for b in bursts:
            b.disk_size = format_bytes_human(sizes[b.id])

    def update_history_status(self, id_list):
        """
        For each burst_id received in the id_list read new status from DB and return a list [id, new_status] pair.
        """
        result = []
        for b_id in id_list:
            burst = dao.get_burst_by_id(b_id)
            # burst.prepare_after_load()
            if burst is not None:
                if burst.status == burst.BURST_RUNNING:
                    running_time = datetime.now() - burst.start_time
                else:
                    running_time = burst.finish_time - burst.start_time
                running_time = format_timedelta(running_time, most_significant2=False)

                if burst.status == burst.BURST_ERROR:
                    msg = 'Check Operations page for error Message'
                else:
                    msg = ''
                result.append([burst.id, burst.status, burst.is_group, msg, running_time])
            else:
                self.logger.debug("Could not find burst with id=" + str(b_id) + ". Might have been deleted by user!!")
        return result

    def stop_burst(self, burst):
        raise NotImplementedError

    def cancel_or_remove_burst(self, burst_id):
        raise NotImplementedError

    @staticmethod
    def update_simulation_fields(burst_id, op_simulation_id, simulation_gid):
        burst = dao.get_burst_by_id(burst_id)
        burst.fk_simulation_id = op_simulation_id
        burst.simulator_gid = simulation_gid.hex
        dao.store_entity(burst)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Dan Pop <dan.pop@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import numpy
import json
from tvb.core.adapters.abcadapter import ABCAdapterForm
from tvb.core.adapters.abcdisplayer import ABCDisplayer
from tvb.basic.logger.builder import get_logger
from tvb.core.entities.model.model_datatype import DataTypeGroup
from tvb.core.entities.model.model_operation import OperationGroup, STATUS_STARTED
from tvb.core.entities.storage import dao
from tvb.core.adapters.exceptions import LaunchException
from tvb.adapters.datatypes.db.mapped_value import DatatypeMeasureIndex
from tvb.core.entities.filters.chain import FilterChain
from tvb.core.neotraits.forms import TraitDataTypeSelectField
from tvb.core.neotraits.view_model import ViewModel, DataTypeGidAttr


class PseIsoModel(object):
    def __init__(self, range1, range2, apriori_data, metrics, datatype_gids):
        self.log = get_logger(self.__class__.__name__)
        # ranges
        all_numbers_range1, self.range1_name, self.range1 = OperationGroup.load_range_interval_numbers(range1)
        all_numbers_range2, self.range2_name, self.range2 = OperationGroup.load_range_interval_numbers(range2)

        # Data from which to interpolate larger 2-D space
        self.apriori_x = self._prepare_axes(self.range1, all_numbers_range1)
        self.apriori_y = self._prepare_axes(self.range2, all_numbers_range2)
        self.apriori_data = apriori_data
        self.datatypes_gids = datatype_gids
        self.metrics = metrics

    @staticmethod
    def pse_filter():
        return FilterChain(fields=[FilterChain.datatype + '.type'], operations=['!='], values=['SimulatorIndex'])

    @classmethod
    def from_db(cls, operation_group_id):
        """
        Collects from db the information about the operation group that is required by the isocline view.
        """
        operations = dao.get_operations_in_group(operation_group_id)
        operation_group = dao.get_operationgroup_by_id(operation_group_id)

        self = cls(operation_group.range1, operation_group.range2, {},
                   PseIsoModel._find_metrics(operations), None)

        self._fill_apriori_data(operations)
        return self

    @staticmethod
    def _find_metrics(operations):
        """ Search for an operation with results. Then get the metrics of the generated data type"""
        dt_measure = None

        for operation in operations:

            if not operation.has_finished:
                raise LaunchException("Can not display until all operations from this range are finished!")

            op_results = dao.get_results_for_operation(operation.id, PseIsoModel.pse_filter())
            if len(op_results):
                datatype = op_results[0]
                if datatype.type == "DatatypeMeasureIndex":
                    # Load proper entity class from DB.
                    dt_measure = dao.get_generic_entity(DatatypeMeasureIndex, datatype.id)[0]
                else:
                    dt_measure = dao.get_generic_entity(DatatypeMeasureIndex, datatype.gid, 'source_gid')
                    if dt_measure:
                        dt_measure = dt_measure[0]
                break

        if dt_measure:
            return json.loads(dt_measure.metrics)
        else:
            raise LaunchException("No datatypes were generated due to simulation errors. Nothing to display.")

    def _fill_apriori_data(self, operations):
        """ Gather apriori data from the operations. Also gather the datatype gid's"""
        for metric in self.metrics:
            self.apriori_data[metric] = numpy.zeros((self.apriori_x.size, self.apriori_y.size))

        # An 2D array of GIDs which is used later to launch overlay for a DataType
        self.datatypes_gids = [[None for _ in self.range2] for _ in self.range1]

        for operation in operations:
            self.log.debug("Gathering data from operation : %s" % operation.id)
            range_values = eval(operation.range_values)
            key_1 = range_values[self.range1_name]
            index_x = self.range1.index(key_1)
            key_2 = range_values[self.range2_name]
            index_y = self.range2.index(key_2)
            if operation.status == STATUS_STARTED:
                raise LaunchException("Not all operations from this range are complete. Cannot view until then.")

            operation_results = dao.get_results_for_operation(operation.id, PseIsoModel.pse_filter())
            if operation_results:
                datatype = operation_results[0]
                self.datatypes_gids[index_x][index_y] = str(datatype.gid)

                if datatype.type == "DatatypeMeasureIndex":
                    measures = dao.get_generic_entity(DatatypeMeasureIndex, datatype.id)
                else:
                    measures = dao.get_generic_entity(DatatypeMeasureIndex, datatype.gid, 'source_gid')
            else:
                self.datatypes_gids[index_x][index_y] = None
                measures = None

            for metric in self.metrics:
                if measures:
                    self.apriori_data[metric][index_x][index_y] = json.loads(measures[0].metrics)[metric]
                else:
                    self.apriori_data[metric][index_x][index_y] = numpy.NaN

    @staticmethod
    def _prepare_axes(original_range_values, is_numbers):
        result = numpy.array(original_range_values)
        if not is_numbers:
            result = numpy.arange(len(original_range_values))
        return result


class IsoclinePSEAdapterModel(ViewModel):
    datatype_group = DataTypeGidAttr(
        linked_datatype=DataTypeGroup,
        label='Datatype Group'
    )


class IsoclinePSEAdapterForm(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None):
        super(IsoclinePSEAdapterForm, self).__init__(prefix, project_id)
        self.datatype_group = TraitDataTypeSelectField(IsoclinePSEAdapterModel.datatype_group, self,
                                                       name='datatype_group', conditions=self.get_filters())

    @staticmethod
    def get_view_model():
        return IsoclinePSEAdapterModel

    @staticmethod
    def get_required_datatype():
        return DataTypeGroup

    @staticmethod
    def get_input_name():
        return 'datatype_group'

    @staticmethod
    def get_filters():
        return FilterChain(fields=[FilterChain.datatype + ".no_of_ranges"], operations=["=="], values=[2])


class IsoclinePSEAdapter(ABCDisplayer):
    """
    Visualization adapter for Parameter Space Exploration.
    Will be used as a generic visualizer, accessible when input entity is DataTypeGroup.
    Will also be used in Burst as a supplementary navigation layer.
    """

    _ui_name = "Isocline Parameter Space Exploration"
    _ui_subsection = "pse_iso"

    def __init__(self):
        ABCDisplayer.__init__(self)
        self.interp_models = {}
        self.nan_indices = {}

    def get_form_class(self):
        return IsoclinePSEAdapterForm

    def get_required_memory_size(self, view_model):
        # type: (IsoclinePSEAdapterModel) -> int
        """
        Return the required memory to run this algorithm.
        """
        # Don't know how much memory is needed.
        return -1

    # TODO: migrate to neotraits
    def burst_preview(self, view_model):
        # type: (IsoclinePSEAdapterModel) -> dict
        """
        Generate the preview for the burst page.
        """
        return self.launch(view_model)

    def get_metric_matrix(self, datatype_group, selected_metric=None):
        self.model = PseIsoModel.from_db(datatype_group.fk_operation_group)
        if selected_metric is None:
            selected_metric = list(self.model.metrics)[0]

        data_matrix = self.model.apriori_data[selected_metric]
        data_matrix = numpy.rot90(data_matrix)
        data_matrix = numpy.flipud(data_matrix)
        matrix_data = ABCDisplayer.dump_with_precision(data_matrix.flat)
        matrix_guids = self.model.datatypes_gids
        matrix_guids = numpy.rot90(matrix_guids)
        matrix_shape = json.dumps(data_matrix.squeeze().shape)
        x_min = self.model.apriori_x[0]
        x_max = self.model.apriori_x[self.model.apriori_x.size - 1]
        y_min = self.model.apriori_y[0]
        y_max = self.model.apriori_y[self.model.apriori_y.size - 1]
        vmin = data_matrix.min()
        vmax = data_matrix.max()
        return dict(matrix_data=matrix_data,
                    matrix_guids=json.dumps(matrix_guids.flatten().tolist()),
                    matrix_shape=matrix_shape,
                    color_metric=selected_metric,
                    x_min=x_min,
                    x_max=x_max,
                    y_min=y_min,
                    y_max=y_max,
                    vmin=vmin,
                    vmax=vmax)

    @staticmethod
    def prepare_node_data(datatype_group):
        if datatype_group is None:
            raise Exception("Selected DataTypeGroup is no longer present in the database. "
                            "It might have been remove or the specified id is not the correct one.")

        operation_group = dao.get_operationgroup_by_id(datatype_group.fk_operation_group)
        operations = dao.get_operations_in_group(operation_group.id)
        node_info_dict = dict()
        for operation_ in operations:
            datatypes = dao.get_results_for_operation(operation_.id)
            if len(datatypes) > 0:
                datatype = datatypes[0]
                node_info_dict[datatype.gid] = dict(operation_id=operation_.id,
                                                    datatype_gid=datatype.gid,
                                                    datatype_type=datatype.type,
                                                    datatype_subject=datatype.subject,
                                                    datatype_invalid=datatype.invalid)
        return node_info_dict

    def launch(self, view_model):
        datatype_group_index = self.load_entity_by_gid(view_model.datatype_group.hex)
        params = self.get_metric_matrix(datatype_group_index)
        params["title"] = self._ui_name
        params["canvasName"] = "Interpolated values for PSE metric: "
        params["xAxisName"] = self.model.range1_name
        params["yAxisName"] = self.model.range2_name
        params["url_base"] = "/burst/explore/get_metric_matrix/" + view_model.datatype_group.hex
        params["node_info_url"] = "/burst/explore/get_node_matrix/" + view_model.datatype_group.hex
        params["available_metrics"] = list(self.model.metrics)
        return self.build_display_result('pse_isocline/view', params,
                                         pages=dict(controlPage="pse_isocline/controls"))



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Andrei Mihai <mihai.andrei@codemart.ro>

"""

import json
import numpy
from six import add_metaclass
from abc import ABCMeta
from tvb.adapters.visualizers.time_series import ABCSpaceDisplayer
from tvb.basic.neotraits.api import Attr
from tvb.core.entities.filters.chain import FilterChain
from tvb.core.adapters.arguments_serialisation import parse_slice, slice_str
from tvb.core.adapters.abcadapter import ABCAdapterForm
from tvb.core.adapters.abcdisplayer import ABCDisplayer
from tvb.adapters.datatypes.db.spectral import DataTypeMatrix
from tvb.core.neotraits.forms import TraitDataTypeSelectField, StrField
from tvb.core.neocom import h5
from tvb.core.neotraits.view_model import ViewModel, DataTypeGidAttr


# TODO: rewrite, necessary to read whole matrix?
def compute_2d_view(matrix, slice_s):
    """
    Create a 2d view of the matrix using the suggested slice
    If the given slice is invalid or fails to produce a 2d array the default is used
    which selects the first 2 dimensions.
    If the matrix is complex the real part is shown
    :param matrix: main input. It can have more then 2D
    :param slice_s: a string representation of a slice
    :return: (a 2d array,  the slice used to make it, is_default_returned)
    """
    default = (slice(None), slice(None)) + tuple(0 for _ in range(matrix.ndim - 2))  # [:,:,0,0,0,0 etc]

    try:
        if slice_s is not None:
            matrix_slice = parse_slice(slice_s)
        else:
            matrix_slice = slice(None)

        m = matrix[matrix_slice]

        if m.ndim > 2:  # the slice did not produce a 2d array, treat as error
            raise ValueError(str(matrix.shape))

    except (IndexError, ValueError):  # if the slice could not be parsed or it failed to produce a 2d array
        matrix_slice = default

    slice_used = slice_str(matrix_slice)
    return matrix[matrix_slice].astype(float), slice_used, matrix_slice == default


@add_metaclass(ABCMeta)
class MappedArraySVGVisualizerMixin(ABCSpaceDisplayer):
    """
    To be mixed in a ABCDisplayer
    """

    def get_required_memory_size(self, datatype):
        input_size = datatype.read_data_shape()
        return numpy.prod(input_size) / input_size[0] * 8.0

    def generate_preview(self, datatype, **kwargs):
        result = self.launch(datatype)
        result["isPreview"] = True
        return result

    @staticmethod
    def compute_raw_matrix_params(matrix):
        """
        Serializes matrix data, shape and stride metadata to json
        """
        matrix_data = ABCDisplayer.dump_with_precision(matrix.flat)
        matrix_shape = json.dumps(matrix.shape)

        return dict(matrix_data=matrix_data,
                    matrix_shape=matrix_shape)

    def compute_params(self, matrix, viewer_title, given_slice=None, labels=None):
        """
        Prepare a 2d matrix to display
        :param labels: optional labels for the matrix
        :param viewer_title: title for the matrix display
        :param matrix: input matrix
        :param given_slice: a string representation of a slice. This slice should cut a 2d view from matrix
        If the matrix is not 2d and the slice will not make it 2d then a default slice is used
        """
        matrix2d, slice_used, is_default_slice = compute_2d_view(matrix, given_slice)

        view_pars = self.compute_raw_matrix_params(matrix2d)
        view_pars.update(original_matrix_shape=str(matrix.shape),
                         show_slice_info=given_slice is not None,
                         given_slice=given_slice,
                         slice_used=slice_used,
                         is_default_slice=is_default_slice,
                         viewer_title=viewer_title,
                         title=viewer_title,
                         matrix_labels=json.dumps(labels))
        return view_pars

    def _extract_labels_and_data_matrix(self, datatype_index):
        """
        If datatype has a source attribute of type TimeSeriesRegion
        then the labels of the associated connectivity are returned.
        Else None
        """
        with h5.h5_file_for_index(datatype_index) as datatype_h5:
            matrix = datatype_h5.array_data[:]

        source_index = self.load_entity_by_gid(datatype_index.source_gid)
        with h5.h5_file_for_index(source_index) as source_h5:
            labels = self.get_space_labels(source_h5)

        return [labels, labels], matrix


class MatrixVisualizerModel(ViewModel):
    datatype = DataTypeGidAttr(
        linked_datatype=DataTypeMatrix,
        label='Array data type'
    )

    slice = Attr(
        field_type=str,
        required=False,
        label='slice indices in numpy syntax'
    )


class MatrixVisualizerForm(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None):
        super(MatrixVisualizerForm, self).__init__(prefix, project_id, False)
        self.datatype = TraitDataTypeSelectField(MatrixVisualizerModel.datatype, self, name='datatype',
                                                 conditions=self.get_filters())
        self.slice = StrField(MatrixVisualizerModel.slice, self, name='slice')

    @staticmethod
    def get_view_model():
        return MatrixVisualizerModel

    @staticmethod
    def get_input_name():
        return 'datatype'

    @staticmethod
    def get_filters():
        return FilterChain(fields=[FilterChain.datatype + '.ndim'], operations=[">="], values=[2])

    @staticmethod
    def get_required_datatype():
        return DataTypeMatrix


class MappedArrayVisualizer(MappedArraySVGVisualizerMixin):
    _ui_name = "Matrix Visualizer"
    _ui_subsection = "matrix"

    def get_form_class(self):
        return MatrixVisualizerForm

    def launch(self, view_model):
        # type: (MatrixVisualizerModel) -> dict
        datatype_index = self.load_entity_by_gid(view_model.datatype.hex)
        with h5.h5_file_for_index(datatype_index) as h5_file:
            matrix = h5_file.array_data.load()

        matrix2d, _, _ = compute_2d_view(matrix, view_model.slice)
        title = datatype_index.display_name + " matrix plot"

        pars = self.compute_params(matrix, title, view_model.slice)
        return self.build_display_result("matrix/svg_view", pars)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Change of DB structure to TVB 1.0.3.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

from sqlalchemy.sql import text
from sqlalchemy import Table, MetaData, Column, Integer, String
from migrate.changeset.schema import create_column
from tvb.core.entities.storage import SA_SESSIONMAKER


COL_1 = Column('_number_of_values', Integer)
COL_2 = Column('_invdx', String)
COL_3 = Column('_xmax', String)
COL_4 = Column('_df', String)
COL_5 = Column('_dx', String)
COL_6 = Column('_xmin', String)
COL_7 = Column('_data', String)


def _prepare_table(meta, table_name):
    """
    Load current DB structure.
    """
    table = Table(table_name, meta, autoload=True)
    for constraint in table._sorted_constraints:
        if not isinstance(constraint.name, str):
            constraint.name = None
    return table


def upgrade(migrate_engine):
    """
    Upgrade operations go here.
    Don't create your own engine; bind migrate_engine to your metadata.
    """
    meta = MetaData(bind=migrate_engine)
    
    table = _prepare_table(meta, 'USER_PREFERENCES')
    table.c.user_id.alter(name='fk_user')
    
    table = _prepare_table(meta, 'BURST_CONFIGURATIONS')
    table.c.project_id.alter(name='fk_project')
    
    table = _prepare_table(meta, 'WORKFLOWS',)
    table.c.project_id.alter(name='fk_project')
    table.c.burst_id.alter(name='fk_burst')
    
    table = _prepare_table(meta, 'WORKFLOW_STEPS')
    table.c.workflow_id.alter(name='fk_workflow')
    table.c.algorithm_id.alter(name='fk_algorithm')
    table.c.resulted_op_id.alter(name='fk_operation')
    
    table = _prepare_table(meta, 'MAPPED_DATATYPE_MEASURE')
    table.c.analyzed_datatype.alter(name='_analyzed_datatype')
    
    ## Fix Lookup Table mapping.
    table = _prepare_table(meta, 'MAPPED_LOOK_UP_TABLE_DATA')
    create_column(COL_1, table)
    create_column(COL_2, table)
    create_column(COL_3, table)
    create_column(COL_4, table)
    create_column(COL_5, table)
    create_column(COL_6, table)
    create_column(COL_7, table)
    session = SA_SESSIONMAKER()
    session.execute(text('DELETE FROM "MAPPED_LOOK_UP_TABLE_DATA";'))
    session.execute(text('insert into "MAPPED_LOOK_UP_TABLE_DATA"(id, _equation, _number_of_values, _invdx, _xmax, _xmin, _df, _dx, _data) '
                         'select id, \'\', _number_of_values, _invdx, _xmax, _xmin, _df, _dx, _data from "MAPPED_NERF_TABLE_DATA";'))
    session.execute(text('insert into "MAPPED_LOOK_UP_TABLE_DATA"(id, _equation, _number_of_values, _invdx, _xmax, _xmin, _df, _dx, _data) '
                         'select id, \'\', _number_of_values, _invdx, _xmax, _xmin, _df, _dx, _data from "MAPPED_PSI_TABLE_DATA";'))
    session.commit()
    session.close()
        
    table = _prepare_table(meta, 'MAPPED_NERF_TABLE_DATA')
    table.drop()
    
    table = _prepare_table(meta, 'MAPPED_PSI_TABLE_DATA')
    table.drop()



def downgrade(migrate_engine):
    """
    Operations to reverse the above upgrade go in this function.
    """
    meta = MetaData(bind=migrate_engine)
    
    table = _prepare_table(meta, 'USER_PREFERENCES')
    table.c.fk_user.alter(name='user_id')
    
    table = _prepare_table(meta, 'BURST_CONFIGURATIONS')
    table.c.fk_project.alter(name='project_id')
    
    table = _prepare_table(meta, 'WORKFLOWS')
    table.c.fk_project.alter(name='project_id')
    table.c.fk_burst.alter(name='burst_id')
    
    table = _prepare_table(meta, 'WORKFLOW_STEPS')
    table.c.fk_workflow.alter(name='workflow_id')
    table.c.fk_algorithm.alter(name='algorithm_id')
    table.c.fk_operation.alter(name='resulted_op_id')
    
    table = _prepare_table(meta, 'MAPPED_DATATYPE_MEASURE')
    table.c._analyzed_datatype.alter(name='analyzed_datatype') 
    
    
        


# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Change of DB structure from TVB version 1.0.8 to 1.1.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

from sqlalchemy import Column, String
from migrate.changeset.schema import create_column, drop_column
from tvb.basic.logger.builder import get_logger
from tvb.core.entities import model

meta = model.Base.metadata
COL_LABELS = Column('_labels_ordering', String, default=0)


def upgrade(migrate_engine):
    """
    Upgrade operations go here.
    Don't create your own engine; bind migrate_engine to your metadata.
    """
    try:
        meta.bind = migrate_engine
        table = meta.tables['MAPPED_CORRELATION_COEFFICIENTS_DATA']
        create_column(COL_LABELS, table)

    except Exception:
        ## This update might fail, in case people are having an older version of TVB
        ## (in which current datatypes does not even exists)
        logger = get_logger(__name__)
        logger.warning("We do not update table MAPPED_CORRELATION_COEFFICIENTS_DATA, because it is not yet created")


def downgrade(migrate_engine):
    """
    Operations to reverse the above upgrade go here.
    """
    try:
        meta.bind = migrate_engine
        table = meta.tables['MAPPED_CORRELATION_COEFFICIENTS_DATA']
        drop_column(COL_LABELS, table)

    except Exception:
        ## This update is not critical. We can run even in case of error at update
        logger = get_logger(__name__)
        logger.warning("We do not update table MAPPED_CORRELATION_COEFFICIENTS_DATA, because it is not yet created")





# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
from tvb.basic.neotraits.api import NArray
from tvb.datatypes.patterns import StimuliRegion, StimuliSurface
from tvb.core.neotraits.h5 import H5File, Reference, DataSet, EquationScalar


class StimuliRegionH5(H5File):

    def __init__(self, path):
        super(StimuliRegionH5, self).__init__(path)
        self.spatial = EquationScalar(StimuliRegion.spatial, self)
        self.temporal = EquationScalar(StimuliRegion.temporal, self)
        self.connectivity = Reference(StimuliRegion.connectivity, self)
        self.weight = DataSet(StimuliRegion.weight, self)

    def store(self, datatype, scalars_only=False):
        super(StimuliRegionH5, self).store(datatype, scalars_only)
        self.connectivity.store(datatype.connectivity)


class StimuliSurfaceH5(H5File):

    def __init__(self, path):
        super(StimuliSurfaceH5, self).__init__(path)
        self.spatial = EquationScalar(StimuliSurface.spatial, self)
        self.temporal = EquationScalar(StimuliSurface.temporal, self)
        self.surface = Reference(StimuliSurface.surface, self)
        self.focal_points_surface = DataSet(NArray(dtype=int), self, name='focal_points_surface')
        self.focal_points_triangles = DataSet(StimuliSurface.focal_points_triangles, self)

    def store(self, datatype, scalars_only=False):
        super(StimuliSurfaceH5, self).store(datatype, scalars_only)
        self.focal_points_surface.store(datatype.focal_points_surface)




# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
DAO operations related to Algorithms and User Operations are defined here.
 
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

from sqlalchemy import or_, and_
from sqlalchemy import func as func
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import case as case_, desc
from tvb.core.entities.model.model_datatype import DataType
from tvb.core.entities.model.model_operation import *
from tvb.core.entities.storage.root_dao import RootDAO, DEFAULT_PAGE_SIZE


class OperationDAO(RootDAO):
    """
    OPERATION RELATED METHODS
    """


    def get_operation_by_id(self, operation_id):
        """Retrieve OPERATION entity for a given Identifier."""

        operation = self.session.query(Operation).filter_by(id=operation_id).one()
        # Load lazy fields:
        operation.user
        operation.project
        operation.operation_group
        operation.algorithm

        return operation


    def try_get_operation_by_id(self, operation_id):
        """
        Try to call self.get_operation_by_id, but when operation was not found, instead of failing, return None.
        This could be called from situations like: stopping & removing op.
        A check for None is compulsory after this call!
        """

        try:
            return self.get_operation_by_id(operation_id)
        except SQLAlchemyError:
            self.logger.exception("Operation not found for ID %s, we will return None" % operation_id)
            return None


    def get_operation_by_gid(self, operation_gid):
        """Retrieve OPERATION entity for a given gid."""
        try:
            operation = self.session.query(Operation).filter_by(gid=operation_gid).one()
            operation.user
            operation.project
            operation.operation_group
            operation.algorithm.algorithm_category
            return operation
        except SQLAlchemyError:
            self.logger.exception("When fetching gid %s" % operation_gid)
            return None

    def get_operation_lazy_by_gid(self, operation_gid):
        """Retrieve OPERATION entity for a given gid."""
        try:
            return self.session.query(Operation).filter_by(gid=operation_gid).one()
        except SQLAlchemyError:
            self.logger.exception("When fetching gid %s" % operation_gid)
            return None


    def get_all_operations_for_uploaders(self, project_id):
        """
        Returns all finished upload operations.
        """
        try:
            result = self.session.query(Operation).join(Algorithm).join(AlgorithmCategory).filter(
                                        AlgorithmCategory.rawinput == True).filter(
                                        Operation.fk_launched_in == project_id).filter(
                                        Operation.status == STATUS_FINISHED).all()
            return result
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return None


    def is_upload_operation(self, operation_gid):
        """
        Returns True only if the operation with the given gid is an upload operation.
        """
        try:
            result = self.session.query(Operation).join(Algorithm).join(AlgorithmCategory
                                        ).filter(AlgorithmCategory.rawinput == True
                                                 ).filter(Operation.gid == operation_gid).count()
            return result > 0
        except SQLAlchemyError:
            return False


    def count_resulted_datatypes(self, operation_id):
        """
        Returns the number of resulted datatypes from the specified operation.
        """
        try:
            result = self.session.query(DataType).filter_by(fk_from_operation=operation_id).count()
            return result
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return None


    def get_operation_process_for_operation(self, operation_id):
        """
        Get the OperationProcessIdentifier for this operation id.
        """
        try:
            result = self.session.query(OperationProcessIdentifier
                                        ).filter(OperationProcessIdentifier.fk_from_operation == operation_id
                                                 ).one()
        except NoResultFound:
            self.logger.debug("No operation process found for operation id=%s." % (str(operation_id),))
            result = None
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            result = None
        return result


    def get_operations_in_group(self, operation_group_id, is_count=False,
                                only_first_operation=False, only_gids=False):
        """
        Retrieve OPERATION entities for a given group.
        """
        result = None
        try:
            query = self.session.query(Operation)
            if only_gids:
                query = self.session.query(Operation.gid)
            query = query.filter_by(fk_operation_group=operation_group_id)
            if is_count:
                result = query.count()
            elif only_first_operation:
                result = query.first()
            else:
                result = query.all()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
        return result


    def compute_disk_size_for_started_ops(self, user_id):
        """ Get all the disk space that should be reserved for the started operations of this user. """
        try:
            expected_hdd_size = self.session.query(func.sum(Operation.estimated_disk_size)
                                                   ).filter(Operation.fk_launched_by == user_id
                                                   ).filter(Operation.status == STATUS_STARTED).scalar()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            expected_hdd_size = 0
        return expected_hdd_size or 0


    def get_filtered_operations(self, project_id, filter_chain, page_start=0,
                                page_size=DEFAULT_PAGE_SIZE, is_count=False):
        """
        :param project_id: current project ID
        :param filter_chain: instance of FilterChain
        :param is_count: when True, return a number, otherwise the list of operation entities

        :return a list of filtered operation in current project, page by page, or the total count for them.
        """
        try:
            select_clause = self.session.query(func.min(Operation.id))
            if not is_count:
                # Do not add select columns in case of COUNT, as they will be ignored anyway
                select_clause = self.session.query(func.min(Operation.id), func.max(Operation.id),
                                                   func.count(Operation.id),
                                                   func.max(Operation.fk_operation_group),
                                                   func.min(Operation.fk_from_algo),
                                                   func.max(Operation.fk_launched_by),
                                                   func.min(Operation.create_date),
                                                   func.min(Operation.start_date),
                                                   func.max(Operation.completion_date),
                                                   func.min(Operation.status),
                                                   func.max(Operation.additional_info),
                                                   func.min(case_([(Operation.visible, 1)], else_=0)),
                                                   func.min(Operation.user_group),
                                                   func.min(Operation.gid))

            query = select_clause.join(Algorithm).join(
                AlgorithmCategory).filter(Operation.fk_launched_in == project_id)

            if filter_chain is not None:
                filter_string = filter_chain.get_sql_filter_equivalent()
                query = query.filter(eval(filter_string))
            query = query.group_by(case_([(Operation.fk_operation_group > 0,
                                           - Operation.fk_operation_group)], else_=Operation.id))

            if is_count:
                return query.count()

            return query.order_by(desc(func.max(Operation.id))).offset(page_start).limit(page_size).all()

        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return 0 if is_count else None


    def get_results_for_operation(self, operation_id, filters=None):
        """
        Retrieve DataTypes entities, resulted after executing an operation.
        """
        try:
            query = self.session.query(DataType
                                       ).filter_by(fk_from_operation=operation_id
                                       ).filter(and_(DataType.type != self.EXCEPTION_DATATYPE_GROUP,
                                                     DataType.type != self.EXCEPTION_DATATYPE_SIMULATION))
            if filters:
                filter_str = filters.get_sql_filter_equivalent()
                if filter_str is not None:
                    query = query.filter(eval(filter_str))
            query = query.order_by(DataType.id)
            result = query.all()
            for dt in result:
                dt.display_name

            return result
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return None


    def get_operations_for_datatype(self, datatype_gid, only_relevant=True, only_in_groups=False):
        """
        Returns all the operations which uses as an input parameter
        the dataType with the specified GID.
        If the flag only_relevant is True than only the relevant operations will be returned.
    
        If only_in_groups is True than this method will return only the operations that are part
        from an operation group, otherwise it will return only the operations that are NOT part of an operation group.
        """
        try:
            query = self.session.query(Operation).filter(
                                Operation.parameters.like('%' + datatype_gid + '%')).join(
                                Algorithm).join(AlgorithmCategory).filter(
                                AlgorithmCategory.display == False)
            query = self._apply_visibility_and_group_filters(query, only_relevant, only_in_groups)
            result = query.all()
            return result
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return None


    def get_operations_for_datatype_group(self, datatype_group_id, only_relevant=True, only_in_groups=False):
        """
        Returns all the operations which uses as an input parameter a datatype from the given DataTypeGroup.
        If the flag only_relevant is True than only the relevant operations will be returned.
    
        If only_in_groups is True than this method will return only the operations that are
        part from an operation group, otherwise it will return only the operations that
        are NOT part of an operation group.
        """
        try:
            query = self.session.query(Operation).filter(
                DataType.fk_datatype_group == datatype_group_id).filter(
                Operation.parameters.like('%' + DataType.gid + '%')).join(
                Algorithm).join(AlgorithmCategory).filter(
                AlgorithmCategory.display == False)
            query = self._apply_visibility_and_group_filters(query, only_relevant, only_in_groups)
            result = query.all()
            return result
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return None

    @staticmethod
    def _apply_visibility_and_group_filters(query, only_relevant, only_in_groups):
        """
        Used for applying filters on the given query.
        """
        if only_relevant:
            query = query.filter(Operation.visible == True)
        if only_in_groups:
            query = query.filter(Operation.fk_operation_group != None)
        else:
            query = query.filter(Operation.fk_operation_group == None)
        return query


    def set_operation_and_group_visibility(self, entity_gid, is_visible, is_operation_group=False):
        """
        Sets the operation visibility.
    
        If 'is_operation_group' is True than this method will change the visibility for all
        the operation from the OperationGroup with the GID field equal to 'entity_gid'.
        """
        try:
            query = self.session.query(Operation)
            if is_operation_group:
                group = self.get_operationgroup_by_gid(entity_gid)
                query = query.filter(Operation.fk_operation_group == group.id)
            else:
                query = query.filter(Operation.gid == entity_gid)
            query.update({"visible": is_visible})
            self.session.commit()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)


    def get_operationgroup_by_gid(self, gid):
        """Retrieve by GID"""
        try:
            result = self.session.query(OperationGroup).filter_by(gid=gid).one()
            return result
        except SQLAlchemyError:
            return None


    def get_operationgroup_by_id(self, op_group_id):
        """Retrieve by ID"""
        try:
            result = self.session.query(OperationGroup).filter_by(id=op_group_id).one()
            return result
        except SQLAlchemyError:
            return None


    def get_operation_numbers(self, proj_id):
        """
        Count total number of operations started for current project.
        """
        stats = self.session.query(Operation.status, func.count(Operation.id)
                                    ).filter_by(fk_launched_in=proj_id
                                    ).group_by(Operation.status).all()
        stats = dict(stats)
        finished = stats.get(STATUS_FINISHED, 0)
        started = stats.get(STATUS_STARTED, 0)
        failed = stats.get(STATUS_ERROR, 0)
        canceled = stats.get(STATUS_CANCELED, 0)
        pending = stats.get(STATUS_PENDING, 0)

        return finished, started, failed, canceled, pending


    #
    # CATEGORY RELATED METHODS
    #

    def get_algorithm_categories(self):
        """Retrieve all existent categories of Algorithms."""
        try:
            categories = self.session.query(AlgorithmCategory).distinct().order_by(
                AlgorithmCategory.displayname).all()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            categories = []
        return categories


    def get_uploader_categories(self):
        """Retrieve categories with raw_input = true"""
        try:
            result = self.session.query(AlgorithmCategory).filter_by(rawinput=True).all()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            result = []
        return result


    def get_raw_categories(self):
        """Retrieve categories with raw_input = true"""
        try:
            result = self.session.query(AlgorithmCategory).filter_by(defaultdatastate='RAW_DATA').all()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            result = []
        return result


    def get_visualisers_categories(self):
        """Retrieve categories with display = true"""
        try:
            result = self.session.query(AlgorithmCategory).filter_by(display=True).all()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            result = []
        return result


    def get_launchable_categories(self, elimin_viewers=False):
        """Retrieve algorithm categories which can be launched on right-click (optionally filter visualizers)"""
        try:
            result = self.session.query(AlgorithmCategory).filter_by(launchable=True)
            if elimin_viewers:
                result = result.filter_by(display=False)
            result = result.all()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            result = []
        return result


    def get_category_by_id(self, categ_id):
        """Retrieve category with given id"""
        try:
            result = self.session.query(AlgorithmCategory).filter_by(id=categ_id).one()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            result = None
        return result


    def filter_category(self, displayname, rawinput, display, launchable, order_nr):
        """Retrieve category with given id"""
        try:
            result = self.session.query(AlgorithmCategory
                                        ).filter_by(displayname=displayname
                                        ).filter_by(rawinput=rawinput).filter_by(display=display
                                        ).filter_by(launchable=launchable).filter_by(order_nr=order_nr).one()
            return result
        except NoResultFound:
            return None


    #
    # ALGORITHM RELATED METHODS
    #

    def get_algorithm_by_id(self, algorithm_id):
        try:
            result = self.session.query(Algorithm).filter_by(id=algorithm_id).one()
            result.algorithm_category
            return result
        except SQLAlchemyError as ex:
            self.logger.exception(ex)
            return None


    def get_algorithm_by_module(self, module_name, class_name):
        try:
            result = self.session.query(Algorithm).filter_by(module=module_name, classname=class_name).one()
            result.algorithm_category
            return result
        except SQLAlchemyError:
            return None


    def get_applicable_adapters(self, compatible_class_names, launch_categ):
        """
        Retrieve a list of algorithms in a given list of categories with a given dataType classes as required input.
        """
        try:
            return self.session.query(Algorithm
                                      ).filter_by(removed=False
                                      ).filter(Algorithm.fk_category.in_(launch_categ)
                                      ).filter(Algorithm.required_datatype.in_(compatible_class_names)
                                      ).order_by(Algorithm.fk_category
                                      ).order_by(Algorithm.group_name).all()
        except SQLAlchemyError:
            self.logger.exception("Could not retrieve applicable Adapters ...")
            return []


    def get_adapters_from_categories(self, categories):
        """
        Retrieve a list of stored adapters in the given categories.
        """
        try:
            return self.session.query(Algorithm
                                      ).filter_by(removed=False
                                      ).filter(Algorithm.fk_category.in_(categories)
                                      ).order_by(Algorithm.group_name
                                      ).order_by(Algorithm.displayname).all()
        except SQLAlchemyError:
            self.logger.exception("Could not retrieve Adapters ...")
            return []


    #
    # RESULT FIGURE RELATED CODE
    #

    def load_figure(self, figure_id):
        """ Load a figure with all it's lazy load fields to have all required 
        info available. """
        try:
            figure = self.session.query(ResultFigure).filter_by(id=figure_id).one()
            figure.project
            figure.operation
            return figure
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return None


    def get_previews(self, project_id, user_id=None, selected_session_name='all_sessions'):
        """
        This method returns a tuple of 2 elements. The first element represents a dictionary
        of form {'$session_name': [list_of_figures]}. This dictionary contains data only for the selected self.session.
        If the selected session is 'all_sessions' than it will contain data for all the sessions.
        The second element of the returned tuple is a dictionary of form
        {'$session_name': $no_of_figures_in_this_session, ...}.
        This dictionary contains information about all the sessions.
    
        selected_session_name - represents the name of the session for which you
                                want to obtain the stored figures.
        """
        try:
            previews_info = self._get_previews_info(project_id, user_id)
            if selected_session_name == 'all_sessions':
                session_names = list(previews_info)
                session_names.sort()
            else:
                session_names = [selected_session_name]

            result = {}
            for session_name in session_names:
                figures_list = self.session.query(ResultFigure
                                                  ).filter_by(fk_in_project=project_id
                                                              ).filter_by(session_name=session_name)
                if user_id is not None:
                    figures_list = figures_list.filter_by(fk_for_user=user_id)

                figures_list = figures_list.order_by(desc(ResultFigure.id)).all()

                # Force loading of project and operation - needed to compute image path
                for figure in figures_list:
                    figure.project
                    figure.operation
                result[session_name] = figures_list
            return result, previews_info
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return {}, {}


    def _get_previews_info(self, project_id, user_id):
        """
        Returns a dictionary of form: {$session_name: $no_of_images_in_this_session, ...}.
        """
        try:
            result = {}
            session_items = self.session.query(ResultFigure.session_name,
                                               func.count(ResultFigure.session_name)
                                               ).filter_by(fk_in_project=project_id)
            if user_id is not None:
                session_items = session_items.filter_by(fk_for_user=user_id)

            session_items = session_items.group_by(ResultFigure.session_name
                                                   ).order_by(ResultFigure.session_name).all()

            for item in session_items:
                result[item[0]] = item[1]
            return result

        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return {}


    def get_figure_count(self, project_id, user_id):
        """
        Used to generate sequential image names.
        """
        try:
            session_items = self.session.query(ResultFigure).filter_by(fk_in_project=project_id)
            if user_id is not None:
                session_items = session_items.filter_by(fk_for_user=user_id)

            return session_items.count()
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return {}


    def get_figures_for_operation(self, operation_id):
        """Retrieve Figure entities, resulted after executing an operation."""
        try:
            result = self.session.query(ResultFigure).filter_by(fk_from_operation=operation_id).all()
            for figure in result:
                figure.project
                figure.operation
            return result
        except SQLAlchemyError as excep:
            self.logger.exception(excep)
            return None



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
import uuid
from tvb.basic.neotraits.api import Attr
from tvb.simulator.simulator import Simulator
from tvb.core.entities.file.simulator.configurations_h5 import SimulatorConfigurationH5
from tvb.core.neotraits.h5 import Reference, Scalar, Json, DataSet


class SimulatorH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(SimulatorH5, self).__init__(path)
        self.connectivity = Reference(Simulator.connectivity, self)
        self.conduction_speed = Scalar(Simulator.conduction_speed, self)
        self.coupling = Reference(Simulator.coupling, self)
        self.surface = Reference(Simulator.surface, self)
        self.stimulus = Reference(Simulator.stimulus, self)
        self.model = Reference(Simulator.model, self)
        self.integrator = Reference(Simulator.integrator, self)
        self.initial_conditions = DataSet(Simulator.initial_conditions, self)
        self.monitors = Json(Simulator.monitors, self)
        self.simulation_length = Scalar(Simulator.simulation_length, self)
        self.simulation_state = Reference(Attr(field_type=uuid.UUID), self, name='simulation_state')

    def store(self, datatype, scalars_only=False, store_references=False):
        # type: (Simulator, bool, bool) -> None
        self.gid.store(datatype.gid)
        self.connectivity.store(datatype.connectivity)
        self.conduction_speed.store(datatype.conduction_speed)
        self.initial_conditions.store(datatype.initial_conditions)
        self.simulation_length.store(datatype.simulation_length)

        integrator_gid = self.store_config_as_reference(datatype.integrator)
        self.integrator.store(integrator_gid)

        coupling_gid = self.store_config_as_reference(datatype.coupling)
        self.coupling.store(coupling_gid)

        model_gid = self.store_config_as_reference(datatype.model)
        self.model.store(model_gid)

        # TODO: handle multiple monitors
        monitor_gid = self.store_config_as_reference(datatype.monitors[0])
        self.monitors.store([monitor_gid.hex])

        if datatype.surface:
            cortex_gid = self.store_config_as_reference(datatype.surface)
            self.surface.store(cortex_gid)

        if datatype.stimulus:
            self.stimulus.store(datatype.stimulus)

        self.type.store(self.get_full_class_name(type(datatype)))

    def load_into(self, datatype):
        # type: (Simulator) -> None
        datatype.conduction_speed = self.conduction_speed.load()
        datatype.initial_conditions = self.initial_conditions.load()
        datatype.simulation_length = self.simulation_length.load()
        datatype.integrator = self.load_from_reference(self.integrator.load())
        datatype.coupling = self.load_from_reference(self.coupling.load())
        datatype.model = self.load_from_reference(self.model.load())
        # TODO: handle multiple monitors
        datatype.monitors = [self.load_from_reference(self.monitors.load()[0])]
        if self.surface.load():
            datatype.surface = self.load_from_reference(self.surface.load())



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

import os
import numpy
from abc import ABCMeta
from scipy import io as scipy_io
from tvb.basic.logger.builder import get_logger
from tvb.core.adapters.abcadapter import ABCSynchronous, ABCAdapterForm
from tvb.core.entities.transient.structure_entities import DataTypeMetaData
from tvb.core.neotraits.forms import StrField
from tvb.core.neotraits.uploader_view_model import UploaderViewModel


class ABCUploaderForm(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None):
        super(ABCUploaderForm, self).__init__(prefix, project_id)
        self.subject_field = StrField(UploaderViewModel.data_subject, self, name='Data_Subject')
        self.temporary_files = []

    @staticmethod
    def get_required_datatype():
        return None

    @staticmethod
    def get_filters():
        return None

    @staticmethod
    def get_input_name():
        return None


class ABCUploader(ABCSynchronous, metaclass=ABCMeta):
    """
    Base class of the uploaders
    """
    LOGGER = get_logger(__name__)

    def _prelaunch(self, operation, uid=None, available_disk_space=0, view_model=None, **kwargs):
        """
        Before going with the usual prelaunch, get from input parameters the 'subject'.
        """

        self.meta_data.update({DataTypeMetaData.KEY_SUBJECT: view_model.data_subject})
        self.generic_attributes.subject = view_model.data_subject

        return ABCSynchronous._prelaunch(self, operation, uid, available_disk_space, view_model, **kwargs)

    def get_required_memory_size(self, view_model):
        """
        Return the required memory to run this algorithm.
        As it is an upload algorithm and we do not have information about data, we can not approximate this.
        """
        return -1

    def get_required_disk_size(self, view_model):
        """
        As it is an upload algorithm and we do not have information about data, we can not approximate this.
        """
        return 0

    @staticmethod
    def read_list_data(full_path, dimensions=None, dtype=numpy.float64, skiprows=0, usecols=None):
        """
        Read numpy.array from a text file or a npy/npz file.
        """
        try:
            if full_path.endswith(".npy") or full_path.endswith(".npz"):
                array_result = numpy.load(full_path)
            else:
                array_result = numpy.loadtxt(full_path, dtype=dtype, skiprows=skiprows, usecols=usecols)
            if dimensions:
                return array_result.reshape(dimensions)
            return array_result
        except ValueError as exc:
            file_ending = os.path.split(full_path)[1]
            exc.args = (exc.args[0] + " In file: " + file_ending,)
            raise

    @staticmethod
    def read_matlab_data(path, matlab_data_name=None):
        """
        Read array from matlab file.
        """
        try:
            matlab_data = scipy_io.matlab.loadmat(path)
        except NotImplementedError:
            ABCUploader.LOGGER.error("Could not read Matlab content from: " + path)
            ABCUploader.LOGGER.error("Matlab files must be saved in a format <= -V7...")
            raise

        try:
            return matlab_data[matlab_data_name]
        except KeyError:
            def double__(n):
                n = str(n)
                return n.startswith('__') and n.endswith('__')

            available = [s for s in matlab_data if not double__(s)]
            raise KeyError("Could not find dataset named %s. Available datasets: %s" % (matlab_data_name, available))

    @staticmethod
    def get_upload_information():
        return NotImplementedError
